#!/usr/bin/perl

use Getopt::Long;
use Image::Size;
use Math::Geometry;

use strict;

my($version) = "2.0";

# argument defaults
my($map_image) = "";
my($author) = "";
my($ent_file) = "";
my($ent_cfg) = "entity.cfg";
my($altitude) = 4000;
my($yaw) = 0;
my($shift_x) = 0;
my($shift_y) = 0;
my($icon_dir) = "entity_icons";
my($html_dir) = ".";
my($map_path) = "overheads";
my($levelshot_path) = "levelshots";
my($icon_path) = "entity_icons";
my($top_page) = "index";
my($prev_page) = "";
my($next_page) = "";
my($no_ctf) = 0;
my($no_cctf) = 0;
my($no_ta) = 0;
my($no_nav) = 0;
my($no_write) = 0;
my($help) = 0;

# image info
my($map_width, $map_height);
my($bspname);

# entity info
my(@entities);
my(%teams, %teams_new);
my($map_name);
my($center_x_gameunits, $center_y_gameunits, $center_z_gameunits);
my($z_index);
my(%prefix_counts);
my(%entity_counts);

# icon info
my(%entity_icons);
my(%non_icon);

# some "constants"
my($NORMAL_OFFSET) = 0;
my($FLAG_OFFSET) = 1;
my($DOCK_BESIDE) = 0;
my($DOCK_BELOW) = 1;


sub InitIcons {
   my ($classname);

   if (!open(ENTCFG, "$ent_cfg")) {
      die "FATAL: Couldn't open entity config file $ent_cfg!";
   }

   while (<ENTCFG>) {
      s/\/\/.*//;
      if (/\S/) {
         chomp;
         if (/([^,]*),([^,]*),([^,]*),([^,]*),([^,]*),([^,]*),([^,]*)/) {
            if (defined($entity_icons{$1})) {
               die "FATAL: $1 defined twice in entity config file $ent_cfg!";
            }
            $entity_icons{$1}{bad_gametypes} = $2;
            if ($3 ne "") {
               $entity_icons{$1}{filename} = "$icon_dir/$3";
            }
            else {
               $entity_icons{$1}{filename} = "";
            }
            $entity_icons{$1}{origin_local_z} = $4;
            $entity_icons{$1}{alignment} = $5;
            $entity_icons{$1}{prefix} = $6;
            $entity_icons{$1}{abbrev} = $7;
            $entity_icons{$1}{loaded} = 0;
            if ($3 eq "") {
               $entity_icons{$1}{loaded} = 1;
               $entity_icons{$1}{width} = -1;
               $entity_icons{$1}{height} = -1;
               $entity_icons{$1}{prefix} = "m";
            }
         }
         else {
            die "FATAL: badly formatted line in entity config file $ent_cfg:\n$_";
         }
      }
   }

   close(ENTCFG);

   $non_icon{width} = -1;
   $non_icon{height} = -1;
   $non_icon{prefix} = "m";
}

sub LoadIcon {
   my($icon_ref) = @_;

   if (!$$icon_ref{loaded}) {
      $$icon_ref{loaded} = 1;
      ($$icon_ref{width}, $$icon_ref{height}) = imgsize($$icon_ref{filename});
      if (!defined($$icon_ref{width})) {
         die "FATAL: Couldn't read size of icon $$icon_ref{filename}!\n";
      }
      $$icon_ref{path} =
         substr($$icon_ref{filename}, rindex($$icon_ref{filename}, "/") + 1);
      $$icon_ref{path} = "$icon_path/".$$icon_ref{path};
   }
}

sub FindEntity {
   my($entity_key, $entity_value, $start_num) = @_;
   my($entity_num);

   if ($start_num < 0) {
      $start_num = 0;
   }
   elsif ($start_num > $#entities) {
      return(-1);
   }

   foreach $entity_num ($start_num .. $#entities) {
      if (defined($entities[$entity_num]{$entity_key}) &&
          ($entities[$entity_num]{$entity_key} eq $entity_value)) {
         return($entity_num);
      }
   }

   return(-1);
}

sub RemoveEntity {
   my($delete_num) = @_;
   my($entity_num);

   foreach $entity_num ($delete_num .. ($#entities - 1)) {
      $entities[$entity_num] = $entities[$entity_num + 1];
   }

   delete($entities[$#entities]);
}

sub CheckEntityKeys {
   my($entity_keys_ref) = @_;
   my($entity_name);

   $entity_name = $$entity_keys_ref{classname};

   if ($entity_name eq "worldspawn") {
      if (defined($$entity_keys_ref{message})) {
         $map_name = $$entity_keys_ref{message};
      }
      else {
         if (!$no_write) {
            print "\nWARNING 00: worldspawn without message key; cannot get map title.\n";
         }
      }
   }
   elsif (($entity_name eq "team_CTF_redflag") ||
          ($entity_name eq "team_CTF_blueflag")) {
      if (!$no_ta) {
         if (!defined($$entity_keys_ref{gametype})) {
            print "\nWARNING 16: $entity_name at <$$entity_keys_ref{origin}> does not have a gametype key; the flag will be visible in the wrong TA gametypes.\n";
         }
         elsif (($$entity_keys_ref{gametype} =~ /obelisk/) ||
                ($$entity_keys_ref{gametype} =~ /harvester/)) {
            print "\nWARNING 17: $entity_name at <$$entity_keys_ref{origin}> is marked to appear in overload and/or harvester; these are incorrect gametypes.\n";
         }
         elsif (($$entity_keys_ref{gametype} !~ /ctf/) ||
                ($$entity_keys_ref{gametype} !~ /oneflag/)) {
            print "\nWARNING 18: $entity_name at <$$entity_keys_ref{origin}> has gametype \"$$entity_keys_ref{gametype}\"; likely this should instead be \"ctf oneflag\", if you plan to have the target flag be in the same position in both of those gametypes.\n";
         }
      }
      if (!defined($$entity_keys_ref{angle})) {
         print "\nWARNING 25: $entity_name at <$$entity_keys_ref{origin}> does not have an \"angle\" key specified; it should (unless you are intentionally assuming angle 0), to set it in a sensible orientation for gametypes that do not rotate the flags.\n";
      }
   }
   elsif ((!$no_ta) && ($entity_name eq "team_CTF_neutralflag")) {
      if (!defined($$entity_keys_ref{gametype})) {
         print "\nWARNING 19: $entity_name at <$$entity_keys_ref{origin}> does not have a gametype key; the flag will be visible in the wrong TA gametypes.\n";
      }
      elsif ($$entity_keys_ref{gametype} ne "oneflag") {
         print "\nWARNING 20: $entity_name at <$$entity_keys_ref{origin}> has gametype \"$$entity_keys_ref{gametype}\"; instead this should be \"oneflag\".\n";
      }
   }
   elsif (($entity_name eq "team_redobelisk") ||
          ($entity_name eq "team_blueobelisk")) {
      if (defined($$entity_keys_ref{gametype})) {
         if (($$entity_keys_ref{gametype} !~ /ctf/) ||
             ($$entity_keys_ref{gametype} !~ /oneflag/) ||
             ($$entity_keys_ref{gametype} !~ /obelisk/) ||
             ($$entity_keys_ref{gametype} !~ /harvester/)) {
            print "\nWARNING 21: $entity_name at <$$entity_keys_ref{origin}> has gametype \"$$entity_keys_ref{gametype}\"; likely this should instead include all gametypes, or the key be omitted, if you plan to have the obelisks be in the same positions as the flags.\n";
         }
      }
   }
   elsif ($entity_name eq "team_neutralobelisk") {
      if (!defined($$entity_keys_ref{gametype})) {
         print "\nWARNING 22: $entity_name at <$$entity_keys_ref{origin}> does not have a gametype key; the flagstand will be visible in the wrong TA gametypes.\n";
      }
      elsif (($$entity_keys_ref{gametype} =~ /obelisk/) ||
             ($$entity_keys_ref{gametype} =~ /ctf/)) {
         print "\nWARNING 23: $entity_name at <$$entity_keys_ref{origin}> is marked to appear in ctf and/or overload; these are incorrect gametypes.\n";
      }
      elsif (($$entity_keys_ref{gametype} !~ /harvester/) ||
             ($$entity_keys_ref{gametype} !~ /oneflag/)) {
         print "\nWARNING 24: $entity_name at <$$entity_keys_ref{origin}> has gametype \"$$entity_keys_ref{gametype}\"; likely this should instead be \"oneflag harvester\", if you plan to have the oneflag neutralflag be in the same position as the harvester generator.\n";
      }
   }
   elsif (($entity_name eq "item_guard") ||
          ($entity_name eq "item_scout") ||
          ($entity_name eq "item_doubler") ||
          ($entity_name eq "item_ammoregen")) {
      if (!defined($$entity_keys_ref{spawnflags})) {
         print "\nWARNING 15: $entity_name at <$$entity_keys_ref{origin}> is not marked for a team.\n";
      }
      else {
         if ($$entity_keys_ref{spawnflags} == 2) {
            $entity_name .= "_red";
         }
         else {
            $entity_name .= "_blue";
         }
      }
   }

   if (!defined($entity_counts{$entity_name})) {
      $entity_counts{$entity_name} = 1;
   }
   else {
      $entity_counts{$entity_name}++;
   }
}

sub CheckEntityCounts {
   if ((!defined($entity_counts{info_player_start})) ||
       ($entity_counts{info_player_start} != 1)) {
      print "\nWARNING 26: There should be 1 and only 1 info_player_start (this is more just a matter of custom than a real problem however).\n";
   }
   if ((!defined($entity_counts{info_player_intermission})) ||
       ($entity_counts{info_player_intermission} != 1)) {
      print "\nWARNING 27: There should be 1 and only 1 info_player_intermission, unless you are manipulating the gametype key to have different intermission points for different gametypes.\n";
   }
   if ((!defined($entity_counts{team_CTF_redflag})) ||
       ($entity_counts{team_CTF_redflag} != 1)) {
      if ($no_ta) {
         print "\nWARNING 28: There should be 1 and only 1 team_CTF_redflag.\n";
      }
      else {
         print "\nWARNING 28: There should be 1 and only 1 team_CTF_redflag, unless you are manipulating the gametype key to place them in different spots for ctf vs. oneflag play.\n";
      }
   }
   if ((!defined($entity_counts{team_CTF_blueflag})) ||
       ($entity_counts{team_CTF_blueflag} != 1)) {
      if ($no_ta) {
         print "\nWARNING 29: There should be 1 and only 1 team_CTF_blueflag.\n";
      }
      else {
         print "\nWARNING 29: There should be 1 and only 1 team_CTF_blueflag, unless you are manipulating the gametype key to place them in different spots for ctf vs. oneflag play.\n";
      }
   }
   if (!$no_ta) {
      if ((!defined($entity_counts{team_CTF_neutralflag})) ||
          ($entity_counts{team_CTF_neutralflag} != 1)) {
         print "\nWARNING 30: There should be 1 and only 1 team_CTF_neutralflag.\n";
      }
      if ((!defined($entity_counts{team_redobelisk})) ||
          ($entity_counts{team_redobelisk} != 1)) {
         print "\nWARNING 31: There should be 1 and only 1 team_redobelisk, unless you are manipulating the gametype key to place the obelisks in different spots than the flags.\n";
      }
      if ((!defined($entity_counts{team_blueobelisk})) ||
          ($entity_counts{team_blueobelisk} != 1)) {
         print "\nWARNING 32: There should be 1 and only 1 team_blueobelisk, unless you are manipulating the gametype key to place the obelisks in different spots than the flags.\n";
      }
      if ((!defined($entity_counts{team_neutralobelisk})) ||
          ($entity_counts{team_neutralobelisk} != 1)) {
         print "\nWARNING 33: There should be 1 and only 1 team_neutralobelisk, unless you are manipulating the gametype key to place the oneflag neutralflag in a different spot than the harvester generator.\n";
      }
   }

   if (!defined($entity_counts{team_CTF_redplayer})) {
      print "\nWARNING 34: There are no team_CTF_redplayer spawnpoints; they are required.\n";
   }
   else {
      if (($entity_counts{team_CTF_redplayer} < 10) ||
          ($entity_counts{team_CTF_redplayer} > 16)) {
         print "\nWARNING 35: There are $entity_counts{team_CTF_redplayer} team_CTF_redplayer spawnpoints; generally 10-16 are recommended.\n";
      }
   }
   if (!defined($entity_counts{team_CTF_blueplayer})) {
      print "\nWARNING 36: There are no team_CTF_blueplayer spawnpoints; they are required.\n";
   }
   else {
      if (($entity_counts{team_CTF_blueplayer} < 10) ||
          ($entity_counts{team_CTF_blueplayer} > 16)) {
         print "\nWARNING 37: There are $entity_counts{team_CTF_blueplayer} team_CTF_blueplayer spawnpoints; generally 10-16 are recommended.\n";
      }
      if (defined($entity_counts{team_CTF_redplayer})) {
         if ($entity_counts{team_CTF_blueplayer} !=
             $entity_counts{team_CTF_redplayer}) {
            print "\nWARNING 38: There are $entity_counts{team_CTF_redplayer} team_CTF_redplayer spawnpoints and $entity_counts{team_CTF_blueplayer} team_CTF_blueplayer spawnpoints; these should be equal, especially if the map is symmetrical.\n";
         }
      }
   }

   if (!defined($entity_counts{team_CTF_redspawn})) {
      print "\nWARNING 39: There are no team_CTF_redspawn respawnpoints; they are required.\n";
   }
   else {
      if (($entity_counts{team_CTF_redspawn} < 10) ||
          ($entity_counts{team_CTF_redspawn} > 16)) {
         print "\nWARNING 40: There are $entity_counts{team_CTF_redspawn} team_CTF_redspawn respawnpoints; generally 10-16 are recommended.\n";
      }
   }
   if (!defined($entity_counts{team_CTF_bluespawn})) {
      print "\nWARNING 41: There are no team_CTF_bluespawn respawnpoints; they are required.\n";
   }
   else {
      if (($entity_counts{team_CTF_bluespawn} < 10) ||
          ($entity_counts{team_CTF_bluespawn} > 16)) {
         print "\nWARNING 42: There are $entity_counts{team_CTF_bluespawn} team_CTF_bluespawn respawnpoints; generally 10-16 are recommended.\n";
      }
      if (defined($entity_counts{team_CTF_redspawn})) {
         if ($entity_counts{team_CTF_bluespawn} !=
             $entity_counts{team_CTF_redspawn}) {
            print "\nWARNING 43: There are $entity_counts{team_CTF_redspawn} team_CTF_redspawn respawnpoints and $entity_counts{team_CTF_bluespawn} team_CTF_bluespawn respawnpoints; these should be equal, especially if the map is symmetrical.\n";
         }
      }
   }

   if (!defined($entity_counts{info_player_deathmatch})) {
      print "\nWARNING 44: There are no info_player_deathmatch spawnpoints; they are recommended.\n";
   }
   else {
      if (($entity_counts{info_player_deathmatch} < 10) ||
          ($entity_counts{info_player_deathmatch} > 16)) {
         print "\nWARNING 45: There are $entity_counts{info_player_deathmatch} info_player_deathmatch spawnpoints; generally 10-16 are recommended, unless you are using them as personal teleporter destinations.\n";
      }
   }

   if (!defined($entity_counts{target_location})) {
      print "\nWARNING 46: There are no target_location entities; they are required.\n";
   }

   if (!$no_ta) {
      if (defined($entity_counts{item_guard_red})) {
         if (!defined($entity_counts{item_guard_blue})) {
            print "\nWARNING 47: The red team has at least 1 guard powerup and the blue team does not.\n";
         }
         else {
            if ($entity_counts{item_guard_red} !=
                $entity_counts{item_guard_blue}) {
               print "\nWARNING 48: Red team item_guard count is $entity_counts{item_guard_red}, blue team item_guard count is $entity_counts{item_guard_blue}; they should be equal, especially if the map is symmetrical.\n";
            }
         }
      }
      if (defined($entity_counts{item_scout_red})) {
         if (!defined($entity_counts{item_scout_blue})) {
            print "\nWARNING 49: The red team has at least 1 scout powerup and the blue team does not.\n";
         }
         else {
            if ($entity_counts{item_scout_red} !=
                $entity_counts{item_scout_blue}) {
               print "\nWARNING 50: Red team item_scout count is $entity_counts{item_scout_red}, blue team item_scout count is $entity_counts{item_scout_blue}; they should be equal, especially if the map is symmetrical.\n";
            }
         }
      }
      if (defined($entity_counts{item_doubler_red})) {
         if (!defined($entity_counts{item_doubler_blue})) {
            print "\nWARNING 51: The red team has at least 1 doubler powerup and the blue team does not.\n";
         }
         else {
            if ($entity_counts{item_doubler_red} !=
                $entity_counts{item_doubler_blue}) {
               print "\nWARNING 52: Red team item_doubler count is $entity_counts{item_doubler_red}, blue team item_doubler count is $entity_counts{item_doubler_blue}; they should be equal, especially if the map is symmetrical.\n";
            }
         }
      }
      if (defined($entity_counts{item_ammoregen_red})) {
         if (!defined($entity_counts{item_ammoregen_blue})) {
            print "\nWARNING 53: The red team has at least 1 ammoregen powerup and the blue team does not.\n";
         }
         else {
            if ($entity_counts{item_ammoregen_red} !=
                $entity_counts{item_ammoregen_blue}) {
               print "\nWARNING 54: Red team item_ammoregen count is $entity_counts{item_ammoregen_red}, blue team item_ammoregen count is $entity_counts{item_ammoregen_blue}; they should be equal, especially if the map is symmetrical.\n";
            }
         }
      }
   }
}

sub InitEntities {
   my($blueflag_x, $blueflag_y, $blueflag_z);
   my($redflag_x, $redflag_y, $redflag_z);
   my($entity_num);
   my($entity_key);
   my($entity_targetnum);
   my($entity_targeternum);
   my($entity_newnum);
   my($keep_entity);
   my(%temp_entity);
   my($team_name);
   my($team_master);
   my($team_population);
   my($nesting_level);
   my($entered);

   # First, grab all entities that have origin or iconloc_target{name}
   # specified.  Also have to keep around entities that have targets or
   # targetnames, temporarily... will delete them soon after targetname
   # processing if they don't have origins.

   if (!open(ENTFILE, "$ent_file")) {
      die "FATAL: Couldn't open entities file $ent_file!";
   }

   @entities = ();
   %teams = ();
   $map_name = "";

   $nesting_level = 0;
   $entered = 0;
   $entity_num = 0;
   %temp_entity = ();

   while (<ENTFILE>) {
      if (/{/) {
         $nesting_level++;
         $entered = 1;
      }
      elsif (/}/) {
         $nesting_level--;
      }

      if ($nesting_level < 0) {
         die "FATAL: Unbalanced brackets in entities file $ent_file!";
      }

      if (($nesting_level == 1) && (/"([^"]*)"\s*"([^"]*)"/)) {
         $temp_entity{$1} = $2;
      }
      elsif (($nesting_level == 0) && $entered) {
         $entered = 0;
         CheckEntityKeys(\%temp_entity);
         if (defined($temp_entity{origin}) ||
             defined($temp_entity{iconloc_target}) ||
             defined($temp_entity{iconloc_targetname}) ||
             defined($temp_entity{target}) ||
             defined($temp_entity{targetname})) {
            foreach $entity_key (keys %temp_entity) {
               $entities[$entity_num]{$entity_key} = $temp_entity{$entity_key};
            }
            if (defined($entities[$entity_num]{origin})) {
               $entities[$entity_num]{original_origin} =
                  $entities[$entity_num]{origin};
            }
            else {
               $entities[$entity_num]{original_origin} = "undefined origin";
            }
            $entity_num++;
         }
         %temp_entity = ();
      }
   }

   close(ENTFILE);

   CheckEntityCounts();

   # Next, look for trigger_teleports pointing at weird things.  Add a
   # misc_teleporter_dest at the target so an icon can be created there.

   foreach $entity_num (0 .. $#entities) {
      if ($entities[$entity_num]{classname} eq "trigger_teleport") {
         if (!defined($entities[$entity_num]{target})) {
            print "\nWARNING 01: trigger_teleport without target.\n";
         }
         else {
            $entity_targetnum =
               FindEntity("targetname",
                          $entities[$entity_num]{target}, 0);
            if ($entity_targetnum == -1) {
               print "\nWARNING 02: trigger_teleport target $entities[$entity_num]{target} not found.\n";
            }
            while ($entity_targetnum != -1) {
               if (($entities[$entity_targetnum]{classname} ne
                    "misc_teleporter_dest") &&
                   !defined($entities[$entity_targetnum]{copied})) {
                  $entity_newnum = $#entities + 1;
                  $entities[$entity_newnum]{classname} =
                     "misc_teleporter_dest";
                  $entities[$entity_newnum]{old_classname} =
                     $entities[$entity_targetnum]{classname};
                  $entities[$entity_newnum]{targetname} =
                     $entities[$entity_targetnum]{targetname};
                  if (defined($entities[$entity_targetnum]{origin})) {
                     $entities[$entity_newnum]{origin} =
                        $entities[$entity_targetnum]{origin};
                     $entities[$entity_newnum]{original_origin} =
                        $entities[$entity_targetnum]{original_origin};
                  }
                  if (defined($entities[$entity_targetnum]{iconloc_target})) {
                     $entities[$entity_newnum]{iconloc_target} =
                        $entities[$entity_targetnum]{iconloc_target};
                  }
                  if (defined($entities[$entity_targetnum]{icon_suppress})) {
                     $entities[$entity_newnum]{icon_suppress} =
                        $entities[$entity_targetnum]{icon_suppress};
                  }
                  $entities[$entity_targetnum]{copied} = 1;
               }
               $entity_targetnum =
                  FindEntity("targetname",
                             $entities[$entity_num]{target},
                             $entity_targetnum + 1);
            }
         }
      }
   }

   # Next, handle targetted entities.  If they are targetted solely by
   # target_give (or by nothing), remove them.

   # Note: Can't use foreach across a list here... the upper bound needs to
   # dynamically change if we yank out entities.

   for ($entity_num = 0; $entity_num <= $#entities; $entity_num++) {
      if (defined($entities[$entity_num]{targetname}) &&
          ((!defined($entities[$entity_num]{icon_suppress})) ||
           (!$entities[$entity_num]{icon_suppress}))) {
         $keep_entity = 0;
         $entity_targeternum =
            FindEntity("target",
                       $entities[$entity_num]{targetname}, 0);
         while ($entity_targeternum != -1) {
            if ($entities[$entity_targeternum]{classname} ne "target_give") {
               $keep_entity = 1;
            }
            $entity_targeternum =
               FindEntity("target",
                          $entities[$entity_num]{targetname},
                          $entity_targeternum + 1);
         }
         if (!$keep_entity) {
            print "\nWARNING 03: $entities[$entity_num]{classname} entity with targetname $entities[$entity_num]{targetname} is targetted only by target_give (or by nothing); it will not be placed on the map.  You may want to add a global comment to the map about this entity.\n";
            RemoveEntity($entity_num);
            $entity_num--;
         }
      }
   }

   # Next, process the uses of iconloc_target.  It is transitive!

   # If an entity still doesn't have an origin after processing, remove it.

   for ($entity_num = 0; $entity_num <= $#entities; $entity_num++) {
      if (defined($entities[$entity_num]{iconloc_target})) {
         $entity_targetnum =
            FindEntity("iconloc_targetname",
                       $entities[$entity_num]{iconloc_target}, 0);
         while (($entity_targetnum != -1) &&
                (defined($entities[$entity_targetnum]{iconloc_target}))) {
            $entity_targetnum =
               FindEntity("iconloc_targetname",
                          $entities[$entity_targetnum]{iconloc_target}, 0);
         }
         if ($entity_targetnum == -1) {
            if (!defined($entities[$entity_num]{old_classname})) {
               print "\nWARNING 04: $entities[$entity_num]{classname} has iconloc_target $entities[$entity_num]{iconloc_target} pointing at nothing (either directly or at end of iconloc_target chain).\n";
            }
         }
         else {
            if (defined($entities[$entity_targetnum]{origin})) {
               $entities[$entity_num]{origin} =
                  $entities[$entity_targetnum]{origin};
            }
            else {
               if (!defined($entities[$entity_num]{old_classname})) {
                  print "\nWARNING 05: $entities[$entity_num]{classname} has iconloc_target $entities[$entity_num]{iconloc_target} pointing at originless entity (either directly or at end of iconloc_target chain).\n";
               }
            }
         }
         if (!defined($entities[$entity_num]{origin})) {
            if ((!defined($entities[$entity_num]{old_classname})) &&
                ((!defined($entities[$entity_num]{icon_suppress})) ||
                 (!$entities[$entity_num]{icon_suppress}))) {
               print "-- Entity cannot be placed on the map.\n";
            }
            RemoveEntity($entity_num);
            $entity_num--;
         }
      }
      elsif (!defined($entities[$entity_num]{origin})) {
         if (($entities[$entity_num]{classname} eq "trigger_teleport") &&
             ((!defined($entities[$entity_num]{icon_suppress})) ||
              (!$entities[$entity_num]{icon_suppress}))) {
            print "\nWARNING 06: trigger_teleport without origin or iconloc_target; it will not be placed on the map.\n";
         }
         RemoveEntity($entity_num);
         $entity_num--;
      }
   }

   # OK!  Now we have a nice set of entities with origins.  Do a little
   # postprocessing of those and also detect teams.

   foreach $entity_num (0 .. $#entities) {
      $entities[$entity_num]{origin} =~ /(\S+)\s+(\S+)\s+(\S+)/;
      $entities[$entity_num]{x} = $1;
      $entities[$entity_num]{y} = $2;
      $entities[$entity_num]{z} = $3;
      delete($entities[$entity_num]{origin});
      if ($entities[$entity_num]{classname} eq "team_CTF_blueflag") {
         $blueflag_x = $entities[$entity_num]{x};
         $blueflag_y = $entities[$entity_num]{y};
         $blueflag_z = $entities[$entity_num]{z};
      }
      elsif ($entities[$entity_num]{classname} eq "team_CTF_redflag") {
         $redflag_x = $entities[$entity_num]{x};
         $redflag_y = $entities[$entity_num]{y};
         $redflag_z = $entities[$entity_num]{z};
      }
      # Don't have enough bitspace to do team analysis of spawns and
      # respawns... presumably those items will never be teamed anyway,
      # but just to make sure, I'll explicitly skip over those.
      if (defined($entities[$entity_num]{team}) &&
          (($entities[$entity_num]{classname} ne "team_CTF_redplayer") &&
           ($entities[$entity_num]{classname} ne "team_CTF_blueplayer") &&
           ($entities[$entity_num]{classname} ne "team_CTF_redspawn") &&
           ($entities[$entity_num]{classname} ne "team_CTF_bluespawn"))) {
         $team_name = $entities[$entity_num]{team};
         if (!defined($teams{$team_name}[0])) {
            $teams{$team_name}[0] = $entity_num;
         }
         else {
            $team_population = 1 + $#{$teams{$team_name}};
            $teams{$team_name}[$team_population] = $entity_num;
            $team_master = $teams{$team_name}[0];
            if (defined($entities[$entity_num]{iconloc_trackmaster}) &&
                $entities[$entity_num]{iconloc_trackmaster}) {
               $entities[$entity_num]{x} =
                  $entities[$team_master]{x};
               $entities[$entity_num]{y} =
                  $entities[$team_master]{y};
               $entities[$entity_num]{z} =
                  $entities[$team_master]{z};
            }
            else {
               if (($entities[$entity_num]{x} !=
                    $entities[$team_master]{x}) ||
                   ($entities[$entity_num]{y} !=
                    $entities[$team_master]{y}) ||
                   ($entities[$entity_num]{z} !=
                    $entities[$team_master]{z})) {
                  print "\nWARNING 07: $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> has a position different from the master for team $team_name ($entities[$team_master]{classname} at <$entities[$team_master]{original_origin}>).\n";
                  print "-- If the teamed entities are supposed to be in the same spot, consider specifying \"iconloc_trackmaster\" \"1\".\n";
               }
            }
         }
      }
   }

   # Set a few global vars and we're done.

   $center_x_gameunits = ($blueflag_x + $redflag_x) / 2.0;
   $center_y_gameunits = ($blueflag_y + $redflag_y) / 2.0;
   $center_z_gameunits = ($blueflag_z + $redflag_z) / 2.0;

   $center_x_gameunits += $shift_x;
   $center_y_gameunits += $shift_y;

   $z_index = 3;
   $prefix_counts{a} = 0;
   $prefix_counts{w} = 0;
   $prefix_counts{p} = 0;
   $prefix_counts{f} = 0;
   $prefix_counts{m} = 0;
   $prefix_counts{r} = 0;
   $prefix_counts{s} = 0;
}

sub TranslateOrigin {
   my($x, $y, $z, $item_height) = @_;
   my($yaw_rad, $pitch_rad);

   # First, get down on the floor.
   $z -= $item_height;

   # Now translate the gameunit space so that the POV is zero.  The
   # POV is $altitude above the centerpoint.
   $x -= $center_x_gameunits;
   $y -= $center_y_gameunits;
   $z -= ($center_z_gameunits + $altitude);

   # Next we need to take into account that Q3 can't really take
   # a shot looking straight down; the best it can do is 87 degrees
   # pitch down.
   # So first we're going to undo the yaw rotation to get the pitch
   # lined up along a sensible axis...
   if ($yaw != 0) {
      $yaw_rad = -deg2rad($yaw);
      ($x, $y, $z) = rotz($x, $y, $z, $yaw_rad);
   }
   # Now rotate around the y axis to get lined up nicely.
   # Tests have shown that even though the view pitch angle is reported
   # at 87 degrees, I have to use a rotation of 2 degrees here rather
   # than 3 degrees to get stuff to line up.  *shrug*
   $pitch_rad = deg2rad(2.0);
   ($x, $y, $z) = roty($x, $y, $z, $pitch_rad);
   # Finally we should rotate back around the Z axis to a yaw of 90
   # degrees, so that the x/y axes of the gameunit space are aligned
   # with the width and height of the image.
   $yaw_rad = deg2rad(90.0);
   ($x, $y, $z) = rotz($x, $y, $z, $yaw_rad);

   # Done!
   return($x, $y, $z);
}

sub GetID {
   my($prefix) = @_;
   my($id);

   $id = $prefix.$prefix_counts{$prefix};
   $prefix_counts{$prefix}++;
   return($id);
}

sub ResetGametypes {
   my($gametypes_ref) = @_;

   $$gametypes_ref{ctf} = 1;
   $$gametypes_ref{cctf} = 1;
   $$gametypes_ref{ta_ctf} = 1;
   $$gametypes_ref{ta_oneflag} = 1;
   $$gametypes_ref{ta_overload} = 1;
   $$gametypes_ref{ta_harvester} = 1;
}

sub GametypesBitfield {
   my($gametypes_ref) = @_;

   return((1 * $$gametypes_ref{ctf}) |
          (2 * $$gametypes_ref{cctf}) |
          (4 * $$gametypes_ref{ta_ctf}) |
          (8 * $$gametypes_ref{ta_oneflag}) |
          (16 * $$gametypes_ref{ta_overload}) |
          (32 * $$gametypes_ref{ta_harvester}));
}

sub GamelistToGametypes {
   my($label_gamelist, $gametypes_ref) = @_;
   my($strpos);

   ResetGametypes($gametypes_ref);
   if (index($label_gamelist, "cctf") != -1) {
      $strpos = index($label_gamelist, "cctf");
      while ($strpos != -1) {
         substr($label_gamelist, $strpos, 4, "");
         $strpos = index($label_gamelist, "cctf");
      }
   }
   else {
      $$gametypes_ref{cctf} = 0;
   }
   if (index($label_gamelist, "ta_ctf") != -1) {
      $strpos = index($label_gamelist, "ta_ctf");
      while ($strpos != -1) {
         substr($label_gamelist, $strpos, 6, "");
         $strpos = index($label_gamelist, "ta_ctf");
      }
   }
   else {
      $$gametypes_ref{ta_ctf} = 0;
   }
   if (!($label_gamelist =~ /ta_oneflag/)) {
      $$gametypes_ref{ta_oneflag} = 0;
   }
   if ((!($label_gamelist =~ /ta_overload/)) &&
       (!($label_gamelist =~ /ta_obelisk/))) {
      $$gametypes_ref{ta_overload} = 0;
   }
   if (!($label_gamelist =~ /ta_harvester/)) {
      $$gametypes_ref{ta_harvester} = 0;
   }
   if (!($label_gamelist =~ /ctf/)) {
      $$gametypes_ref{ctf} = 0;
   }
}

sub PlaceLabel {
   my($icon_ref, $prefix, $label_text, $x_pos, $y_pos, $z_pos, $gametypes_bitfield, $docking, $abs_x, $abs_y) = @_;
   my($id);

   if ($no_write || !$gametypes_bitfield) {
      return;
   }

   $id = GetID($prefix);

   if (defined($docking) && ($docking == $DOCK_BELOW)) {
      if (!$abs_y) {
         $y_pos += ($$icon_ref{height} + 1);
      }
   }
   else {
      if (!$abs_x) {
         $x_pos += ($$icon_ref{width} + 1);
      }
   }

   print HTMLFILE "<div id=\"$id\" style=\"position: absolute; left: $x_pos";
   print HTMLFILE "px; top: $y_pos";
   print HTMLFILE "px; z-index: $z_pos; visibility: hidden; background-color: white\">";
   print HTMLFILE "<img src=\"$icon_path/transpixel.gif\" width=1 height=1 alt=\"\" id=\"$id";
   print HTMLFILE "Img\" name=\"$gametypes_bitfield\">$label_text";
   print HTMLFILE "<img src=\"$icon_path/transpixel.gif\" width=1 height=1 alt=\"\"></div>\n";
}

sub PlaceIcon {
   my($icon_ref, $gametypes_ref, $x, $y, $z) = @_;
   my($item_height);
   my($offset);
   my($x_gameunits, $y_gameunits, $z_gameunits);
   my($x_page_offset, $y_page_offset);
   my($gametypes_bitfield);
   my($id);

   $item_height = $$icon_ref{origin_local_z};
   $offset = $$icon_ref{alignment};

   if ($no_cctf) {
      $$gametypes_ref{cctf} = 0;
   }
   if ($no_ta) {
      $$gametypes_ref{ta_ctf} = 0;
      $$gametypes_ref{ta_oneflag} = 0;
      $$gametypes_ref{ta_overload} = 0;
      $$gametypes_ref{ta_harvester} = 0;
   }

# FIXME: The projection may be different for aspect ratios other than 4:3... test it.
   ($x_gameunits, $y_gameunits, $z_gameunits) =
      TranslateOrigin($x, $y, $z, $item_height);
   $x_page_offset =
      ($map_width / 2.0) +
      (($x_gameunits * $map_width) / (-$z_gameunits * 2.0));
   $y_page_offset =
      ($map_height / 2.0) -
      (($y_gameunits * $map_width) / (-$z_gameunits * 2.0));
   if (defined($offset) && ($offset == $FLAG_OFFSET)) {
      $x_page_offset = int($x_page_offset + 1.5 - $$icon_ref{width});
      $y_page_offset = int($y_page_offset + 1.5 - $$icon_ref{height});
   }
   else {
      $x_page_offset = int($x_page_offset + 0.5 - ($$icon_ref{width} / 2.0));
      $y_page_offset = int($y_page_offset + 0.5 - ($$icon_ref{height} / 2.0));
   }

   $gametypes_bitfield = GametypesBitfield($gametypes_ref);

   if ($gametypes_bitfield && !$no_write) {
      $id = GetID($$icon_ref{prefix});

      print HTMLFILE "<div id=\"$id\" style=\"position: absolute; left: ";
      print HTMLFILE $x_page_offset;
      print HTMLFILE "px; top: ";
      print HTMLFILE $y_page_offset;
      if ($$icon_ref{prefix} eq "m") {
         print HTMLFILE "px; z-index: 2; visibility: hidden\"><img id=\"$id";
      }
      else {
         print HTMLFILE "px; z-index: $z_index; visibility: hidden\"><img id=\"$id";
         $z_index++;
      }
      print HTMLFILE "Img\" name=\"$gametypes_bitfield\" src=\"";
      print HTMLFILE $$icon_ref{path};
      print HTMLFILE "\" width=".$$icon_ref{width};
      print HTMLFILE " height=".$$icon_ref{height}." alt=\"";
      print HTMLFILE $$icon_ref{abbrev}."\"></div>\n";
   }

   return($x_page_offset, $y_page_offset, $gametypes_bitfield);
}

sub NonIcon {
   my($gametypes_ref, $x, $y, $z) = @_;
   my($x_gameunits, $y_gameunits, $z_gameunits);
   my($x_page_offset, $y_page_offset);
   my($gametypes_bitfield);

   if ($no_cctf) {
      $$gametypes_ref{cctf} = 0;
   }
   if ($no_ta) {
      $$gametypes_ref{ta_ctf} = 0;
      $$gametypes_ref{ta_oneflag} = 0;
      $$gametypes_ref{ta_overload} = 0;
      $$gametypes_ref{ta_harvester} = 0;
   }

   ($x_gameunits, $y_gameunits, $z_gameunits) =
      TranslateOrigin($x, $y, $z, 0);
   $x_page_offset =
      ($map_width / 2.0) +
      (($x_gameunits * $map_width) / (-$z_gameunits * 2.0));
   $y_page_offset =
      ($map_height / 2.0) -
      (($y_gameunits * $map_height) / (-$z_gameunits * 1.5));
   $x_page_offset = int($x_page_offset + 0.5);
   $y_page_offset = int($y_page_offset + 0.5);

   $gametypes_bitfield = GametypesBitfield($gametypes_ref);

   return($x_page_offset, $y_page_offset, $gametypes_bitfield);
}

sub ZSort {
   if (($entities[$a]{classname} ne "team_CTF_redflag") &&
       ($entities[$a]{classname} ne "team_CTF_blueflag") &&
       ($entities[$a]{classname} ne "team_redobelisk") &&
       ($entities[$a]{classname} ne "team_blueobelisk")) {
      if (($entities[$b]{classname} ne "team_CTF_redflag") &&
          ($entities[$b]{classname} ne "team_CTF_blueflag") &&
          ($entities[$b]{classname} ne "team_redobelisk") &&
          ($entities[$b]{classname} ne "team_blueobelisk")) {
         # Neither is a flag entity; sort strictly by Z
         return($entities[$a]{z} <=> $entities[$b]{z});
      }
      else {
         # $b is a flag entity and $a is not; return $a before $b
         return(-1);
      }
   }
   else {
      if (($entities[$b]{classname} ne "team_CTF_redflag") &&
          ($entities[$b]{classname} ne "team_CTF_blueflag") &&
          ($entities[$b]{classname} ne "team_redobelisk") &&
          ($entities[$b]{classname} ne "team_blueobelisk")) {
         # $a is a flag entity and $b is not; return $b before $a
         return(1);
      }
      else {
         # Both are flag entities; sort strictly by Z
         return($entities[$a]{z} <=> $entities[$b]{z});
      }
   }
}

sub PlaceEntities {
   my($classname, $x, $y, $z);
   my($entity_num);
   my(%gametypes);
   my($bad_gametype);
   my($x_page_offset, $y_page_offset, $gametypes_bitfield);
   my($icon_ref);
   my($entity_key);
   my($got_label);
   my($suffix);
   my($label_comment);
   my($label_docking);
   my($label_gamelist);
   my($label_typelist);
   my($prefix);
   my($abs_x, $abs_y);

   ResetGametypes(\%gametypes);

   foreach $entity_num (sort ZSort (0 .. $#entities)) {

      if (defined($entities[$entity_num]{icon_suppress}) &&
          $entities[$entity_num]{icon_suppress}) {
         next;
      }

      if (defined($entities[$entity_num]{notq3a}) &&
          $entities[$entity_num]{notq3a}) {
         $gametypes{ctf} = 0;
         $gametypes{cctf} = 0;
      }
      if (defined($entities[$entity_num]{notta}) &&
          $entities[$entity_num]{notta}) {
         $gametypes{ta_ctf} = 0;
         $gametypes{ta_oneflag} = 0;
         $gametypes{ta_overload} = 0;
         $gametypes{ta_harvester} = 0;
      }
      if (defined($entities[$entity_num]{notcctf}) &&
          $entities[$entity_num]{notcctf}) {
         $gametypes{cctf} = 0;
      }
      else {
         if (defined($entities[$entity_num]{notq3ayescctf}) &&
             $entities[$entity_num]{notq3ayescctf}) {
            $gametypes{cctf} = 1;
         }
      }
      if (defined($entities[$entity_num]{notteam}) &&
          $entities[$entity_num]{notteam}) {
         $gametypes{ctf} = 0;
         $gametypes{cctf} = 0;
         $gametypes{ta_ctf} = 0;
         $gametypes{ta_oneflag} = 0;
         $gametypes{ta_overload} = 0;
         $gametypes{ta_harvester} = 0;
      }
      if (defined($entities[$entity_num]{gametype})) {
         if (!($entities[$entity_num]{gametype} =~ /ctf/)) {
            $gametypes{ctf} = 0;
            $gametypes{cctf} = 0;
            $gametypes{ta_ctf} = 0;
         }
         if (!($entities[$entity_num]{gametype} =~ /oneflag/)) {
            $gametypes{ta_oneflag} = 0;
         }
         if (!($entities[$entity_num]{gametype} =~ /obelisk/)) {
            $gametypes{ta_overload} = 0;
         }
         if (!($entities[$entity_num]{gametype} =~ /harvester/)) {
            $gametypes{ta_harvester} = 0;
         }
      }

      $classname = $entities[$entity_num]{classname};
      $x = $entities[$entity_num]{x};
      $y = $entities[$entity_num]{y};
      $z = $entities[$entity_num]{z};

      if (defined($entity_icons{$classname})) {
         $icon_ref = \%{$entity_icons{$classname}};
         foreach $bad_gametype (split(/\s+/, $$icon_ref{bad_gametypes})) {
            $gametypes{$bad_gametype} = 0;
         }
         if ($no_cctf || $no_ta) {
            if (!$no_cctf) {
               if (($gametypes{ta_ctf} || $gametypes{ta_oneflag} || $gametypes{ta_overload} || $gametypes{ta_harvester}) &&
                   ((!$gametypes{ctf}) && (!$gametypes{cctf}))) {
                  print "\nWARNING 57: You specified --notta, but $classname at <$entities[$entity_num]{original_origin}> is present only in TA.\n";
               }
            }
            elsif (!$no_ta) {
               if ($gametypes{cctf} &&
                   ((!$gametypes{ctf}) && (!$gametypes{ta_ctf}) && (!$gametypes{ta_oneflag}) && (!$gametypes{ta_overload}) && (!$gametypes{ta_harvester}))) {
                  print "\nWARNING 58: You specified --notcctf, but $classname at <$entities[$entity_num]{original_origin}> is present only in CCTF.\n";
               }
            }
            else {
               if (($gametypes{cctf} || $gametypes{ta_ctf} || $gametypes{ta_oneflag} || $gametypes{ta_overload} || $gametypes{ta_harvester}) &&
                   (!$gametypes{ctf})) {
                  print "\nWARNING 59: You specified --notcctf and --notta, but $classname at <$entities[$entity_num]{original_origin}> is not present in normal CTF.\n";
               }
            }
         }
         LoadIcon($icon_ref);
         if ($$icon_ref{width} > 0) {
            ($x_page_offset, $y_page_offset, $gametypes_bitfield) =
               PlaceIcon($icon_ref, \%gametypes, $x, $y, $z);
               if ($classname eq "trigger_teleport") {
                  $got_label = 0;
                  foreach $entity_key (keys %{$entities[$entity_num]}) {
                     if ($entity_key =~ /^label_comment(.*)/) {
                        $got_label = 1;
                        last;
                     }
                  }
                  if (!$got_label) {
                     print "\nWARNING 08: trigger_teleport at <$entities[$entity_num]{original_origin}> without label.\n";
                  }
               }
               elsif ($classname eq "misc_teleporter_dest") {
                  $got_label = 0;
                  foreach $entity_key (keys %{$entities[$entity_num]}) {
                     if ($entity_key =~ /^label_comment(.*)/) {
                        $got_label = 1;
                        last;
                     }
                  }
                  if (!$got_label) {
                     if (defined($entities[$entity_num]{old_classname})) {
                        print "\nWARNING 09: $entities[$entity_num]{old_classname} at <$entities[$entity_num]{original_origin}> is used as a teleporter destination and may not have a label relevant to that.\n";
                     }
                     else {
                        print "\nWARNING 10: misc_teleporter_dest at <$entities[$entity_num]{original_origin}> without label.\n";
                     }
                  }
               }
         }
         else {
            ($x_page_offset, $y_page_offset, $gametypes_bitfield) =
               NonIcon(\%gametypes, $x, $y, $z);
         }
      }
      else {
         $icon_ref = \%non_icon;
         ($x_page_offset, $y_page_offset, $gametypes_bitfield) =
            NonIcon(\%gametypes, $x, $y, $z);
      }


      $entities[$entity_num]{type} = $$icon_ref{prefix};
      $entities[$entity_num]{gametypes_bitfield} = $gametypes_bitfield;
      $entities[$entity_num]{gametypes_bitfield_scratch} = $gametypes_bitfield;

      if (($$icon_ref{width} > 0) &&
          defined($entities[$entity_num]{targetname}) &&
          ($entities[$entity_num]{classname} ne "trigger_teleport") &&
          ($entities[$entity_num]{classname} ne "misc_teleporter_dest")) {
            print "\nWARNING 11: $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> has targetname $entities[$entity_num]{targetname}; you may want to add a global comment to the map about what triggers this entity.\n";
      }

      if (($$icon_ref{width} > 0) &&
          defined($entities[$entity_num]{wait})) {
            print "\nWARNING 55: $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> has a specified wait value of $entities[$entity_num]{wait}; you may want to add a global comment to the map about this if it is unusual.\n";
      }

      if (($$icon_ref{width} > 0) &&
          defined($entities[$entity_num]{random})) {
            print "\nWARNING 56: $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> has a randomized factor of $entities[$entity_num]{random} in its respawn delay; you may want to add a global comment to the map about this.\n";
      }

      foreach $entity_key (keys %{$entities[$entity_num]}) {
         if ($entity_key =~ /^label_comment(.*)/) {
            $suffix = $1;
            $label_comment = $entities[$entity_num]{"label_comment$suffix"};
            $label_docking = $DOCK_BESIDE;
            if (defined($entities[$entity_num]{"label_docking$suffix"})) {
               if ($entities[$entity_num]{"label_docking$suffix"} eq "below") {
                  $label_docking = $DOCK_BELOW;
               }
            }
            if (defined($entities[$entity_num]{"label_gamelist$suffix"})) {
               $label_gamelist =
                  $entities[$entity_num]{"label_gamelist$suffix"};
               GamelistToGametypes($label_gamelist, \%gametypes);
               $gametypes_bitfield = GametypesBitfield(\%gametypes);
            }
            if (defined($entities[$entity_num]{"label_typelist$suffix"})) {
               $label_typelist =
                  $entities[$entity_num]{"label_typelist$suffix"};
            }
            else {
               $label_typelist = $$icon_ref{prefix};
            }
            $abs_x = 0;
            $abs_y = 0;
            if (defined($entities[$entity_num]{"label_x$suffix"})) {
               $x_page_offset = $entities[$entity_num]{"label_x$suffix"};
               $abs_x = 1;
            }
            if (defined($entities[$entity_num]{"label_y$suffix"})) {
               $y_page_offset = $entities[$entity_num]{"label_y$suffix"};
               $abs_y = 1;
            }
            foreach $prefix (keys %prefix_counts) {
               if ($label_typelist =~ /$prefix/) {
                  if ($prefix eq "m") {
                     PlaceLabel($icon_ref, $prefix, $label_comment,
                                $x_page_offset, $y_page_offset, 2,
                                $gametypes_bitfield, $label_docking,
                                $abs_x, $abs_y);
                  }
                  else {
                     PlaceLabel($icon_ref, $prefix, $label_comment,
                                $x_page_offset, $y_page_offset, $z_index,
                                $gametypes_bitfield, $label_docking,
                                $abs_x, $abs_y);
                     $z_index++;
                  }
               }
            }
            $entities[$entity_num]{"label_gametypes_bitfield$suffix"} =
               $gametypes_bitfield;
            $entities[$entity_num]{"label_typelist$suffix"} = $label_typelist;
         }
      }

      ResetGametypes(\%gametypes);
   }
}

sub GametypesStrings {
   my($gametypes_bitfield) = @_;
   my($strings);

   $strings = "";

   if ($gametypes_bitfield & 1) {
      $strings .= " ctf";
   }
   if ($gametypes_bitfield & 2) {
      $strings .= " cctf";
   }
   if ($gametypes_bitfield & 4) {
      $strings .= " ta_ctf";
   }
   if ($gametypes_bitfield & 8) {
      $strings .= " ta_oneflag";
   }
   if ($gametypes_bitfield & 16) {
      $strings .= " ta_overload";
   }
   if ($gametypes_bitfield & 32) {
      $strings .= " ta_harvester";
   }

   return($strings);
}

# spawnpoint and respawnpoint item types not included here... presumably
# they don't need to be, which is good since I don't have the bits to
# spare without going to 64-bit data types.
sub CoverageBitfield {
   my($gametypes_bitfield, $typelist) = @_;
   my($coverage_bitfield);

   $coverage_bitfield = 0;

   if ($typelist =~ /a/) {
      $coverage_bitfield |= $gametypes_bitfield;
   }
   if ($typelist =~ /w/) {
      $coverage_bitfield |= ($gametypes_bitfield << 6);
   }
   if ($typelist =~ /p/) {
      $coverage_bitfield |= ($gametypes_bitfield << 12);
   }
   if ($typelist =~ /f/) {
      $coverage_bitfield |= ($gametypes_bitfield << 18);
   }
   if ($typelist =~ /m/) {
      $coverage_bitfield |= ($gametypes_bitfield << 24);
   }

   return($coverage_bitfield);
}

sub CoverageStrings {
   my($coverage_bitfield) = @_;
   my($strings);

   $strings = "";

   if ($coverage_bitfield & 0x3F) {
      $strings .= "\n(a)";
      $strings .= GametypesStrings($coverage_bitfield & 0x3F);
   }
   if ($coverage_bitfield & (0x3F << 6)) {
      $strings .= "\n(w)";
      $strings .= GametypesStrings(($coverage_bitfield & (0x3F << 6)) >> 6);
   }
   if ($coverage_bitfield & (0x3F << 12)) {
      $strings .= "\n(p)";
      $strings .= GametypesStrings(($coverage_bitfield & (0x3F << 12)) >> 12);
   }
   if ($coverage_bitfield & (0x3F << 18)) {
      $strings .= "\n(f)";
      $strings .= GametypesStrings(($coverage_bitfield & (0x3F << 18)) >> 18);
   }
   if ($coverage_bitfield & (0x3F << 24)) {
      $strings .= "\n(m)";
      $strings .= GametypesStrings(($coverage_bitfield & (0x3F << 24)) >> 24);
   }

   return($strings);
}

sub RemoveTeamMember {
   my($team_name, $delete_num) = @_;
   my($member_num);

   foreach $member_num ($delete_num .. ($#{$teams{$team_name}} - 1)) {
      $teams{$team_name}[$member_num] = $teams{$team_name}[$member_num + 1];
   }

   delete($teams{$team_name}[$#{$teams{$team_name}}]);
}

sub DisplayTeam {
   my($team_name) = @_;
   my($member_num);
   my($entity_num);

   print STDOUT "\nTeam name: $team_name\n";

   for ($member_num = 0; $member_num <= $#{$teams{$team_name}}; $member_num++) {
      $entity_num = $teams{$team_name}[$member_num];
      if (!$entities[$entity_num]{gametypes_bitfield}) {
         RemoveTeamMember($team_name, $member_num);
         $member_num--;
      }
      else {
         print STDOUT "   $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> -- ";
         print STDOUT "(".$entities[$entity_num]{type}.")";
         print STDOUT GametypesStrings($entities[$entity_num]{gametypes_bitfield});
         print STDOUT "\n";
      }
   }
}

sub TeamEntsForGametypes {
   my($team_name, $gametypes_mask) = @_;
   my($member_num);
   my($entity_num);
   my($gametypes_bitfield);
   my($valid_count);

   $valid_count = 0;

   foreach $member_num (0 .. $#{$teams{$team_name}}) {
      $entity_num = $teams{$team_name}[$member_num];
      $gametypes_bitfield = $entities[$entity_num]{gametypes_bitfield_scratch};
      if (($gametypes_bitfield & $gametypes_mask) == $gametypes_mask) {
         $valid_count++;
      }
   }

   return($valid_count);
}

sub StoreCoverageForTeamLabels {
   my($team_name) = @_;
   my($member_num);
   my($entity_num);
   my($entity_type);
   my($entity_key);
   my($suffix);
   my($coverage_bitfield);
   my($total_coverage);
   my($redundant);

   $total_coverage = 0;

   foreach $member_num (0 .. $#{$teams{$team_name}}) {
      $entity_num = $teams{$team_name}[$member_num];
      $entity_type = $entities[$entity_num]{type};
      foreach $entity_key (keys %{$entities[$entity_num]}) {
         if ($entity_key =~ /^label_comment(.*)/) {
            $suffix = $1;
            $coverage_bitfield = CoverageBitfield($entities[$entity_num]{"label_gametypes_bitfield$suffix"}, $entities[$entity_num]{"label_typelist$suffix"});
            $entities[$entity_num]{"label_coverage_bitfield$suffix"} =
               $coverage_bitfield;
            $entities[$entity_num]{"label_coverage_bitfield_scratch$suffix"} =
               $coverage_bitfield;
            $suffix = " ".$suffix;
            if ($total_coverage & $coverage_bitfield) {
               $redundant =
                  CoverageStrings($total_coverage & $coverage_bitfield);
               print "\nWARNING 12: label$suffix for $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> redundantly covers teamed entities for:$redundant\n";
            }
            $total_coverage |= $coverage_bitfield;
         }
      }
   }
}

sub SplitTeam {
   my($team_name) = @_;
   my($best_count);
   my($test_mask);
   my($ent_count);
   my($gametypes_mask);
   my($team_newname);
   my($member_num);
   my($entity_num);
   my($gametypes_bitfield);

   # Find the set of gametypes that applies to the most entities in the
   # team... if a tie, pick the larger set of gametypes.
   $best_count = 0;
   for ($test_mask = 63; $test_mask > 0; $test_mask--) {
      $ent_count = TeamEntsForGametypes($team_name, $test_mask);
      if ($ent_count > $best_count) {
         $best_count = $ent_count;
         $gametypes_mask = $test_mask;
      }
   }

   # If there are no entities left in the team, bail.
   if ($best_count == 0) {
      return;
   }

   # Make a new team for that set of gametypes.
   $team_newname = $team_name."_".$gametypes_mask;
   $teams_new{$team_newname}{gametypes_bitfield} = $gametypes_mask;
   foreach $member_num (0 .. $#{$teams{$team_name}}) {
      $entity_num = $teams{$team_name}[$member_num];
      $gametypes_bitfield = $entities[$entity_num]{gametypes_bitfield_scratch};
      if (($gametypes_bitfield & $gametypes_mask) == $gametypes_mask) {
         $teams_new{$team_newname}{members}[1 + $#{$teams_new{$team_newname}{members}}] = $entity_num;
         $entities[$entity_num]{gametypes_bitfield_scratch} &= ~$gametypes_mask;
      }
   }

   # If it is a "team of one", remove it.
   if ($best_count == 1) {
      delete($teams_new{$team_newname});
   }

   # Try to do it again.
   SplitTeam($team_name);
}
 
sub CoverageForSubteam {
   my($team_newname) = @_;
   my($total_coverage);
   my($member_num);
   my($entity_num);

   $total_coverage = 0;

   foreach $member_num (0 .. $#{$teams_new{$team_newname}{members}}) {
      $entity_num = $teams_new{$team_newname}{members}[$member_num];
      $total_coverage |= CoverageBitfield($teams_new{$team_newname}{gametypes_bitfield}, $entities[$entity_num]{type});
   }

   return($total_coverage);
}

sub CheckSubteamLabelCoverage {
   my($team_name, $team_newname, $coverage_bitfield) = @_;
   my($member_num);
   my($entity_num);
   my($entity_type);
   my($entity_key);
   my($suffix);
   my($label_coverage_bitfield);
   my($uncovered_bitfield);
   my($uncovered);

   $uncovered_bitfield = $coverage_bitfield;

   foreach $member_num (0 .. $#{$teams{$team_name}}) {
      $entity_num = $teams{$team_name}[$member_num];
      $entity_type = $entities[$entity_num]{type};
      foreach $entity_key (keys %{$entities[$entity_num]}) {
         if ($entity_key =~ /^label_comment(.*)/) {
            $suffix = $1;
            $label_coverage_bitfield =
               $entities[$entity_num]{"label_coverage_bitfield$suffix"};
            $uncovered_bitfield &= ~$label_coverage_bitfield;
            $entities[$entity_num]{"label_coverage_bitfield_scratch$suffix"} &=
               ~$coverage_bitfield;
         }
      }
   }

   if ($uncovered_bitfield) {
      $uncovered = CoverageStrings($uncovered_bitfield);
      print "\nWARNING 13: Unlabelled team items\n";
      print "entities:";
      foreach $member_num (0 .. $#{$teams_new{$team_newname}{members}}) {
         $entity_num = $teams_new{$team_newname}{members}[$member_num];
         print " $entities[$entity_num]{classname}";
      }
      print "\nfor:$uncovered\n";
   }
}

sub CheckDanglingLabels {
   my($team_name) = @_;
   my($member_num);
   my($entity_num);
   my($entity_type);
   my($entity_key);
   my($suffix);
   my($label_unused_coverage_bitfield);
   my($unused);

   foreach $member_num (0 .. $#{$teams{$team_name}}) {
      $entity_num = $teams{$team_name}[$member_num];
      $entity_type = $entities[$entity_num]{type};
      foreach $entity_key (keys %{$entities[$entity_num]}) {
         if ($entity_key =~ /^label_comment(.*)/) {
            $suffix = $1;
            $label_unused_coverage_bitfield =
               $entities[$entity_num]{"label_coverage_bitfield_scratch$suffix"};
            if ($label_unused_coverage_bitfield) {
               $unused = CoverageStrings($label_unused_coverage_bitfield);
               print "\nWARNING 14: label$suffix for $entities[$entity_num]{classname} at <$entities[$entity_num]{original_origin}> displays when teamed entities do not for:$unused\n";
            }
         }
      }
   }
}

sub AnalyzeTeams {
   my($team_name);
   my($team_newname);
   my($coverage_bitfield);

   %teams_new = ();

   print STDOUT "\n\n*** item teams analysis ***\n";
   foreach $team_name (keys %teams) {
      # Print out basic team information.
      DisplayTeam($team_name);
      # Determine gametypes+enttypes coverage of team labels.
      StoreCoverageForTeamLabels($team_name);
      # Make subteams that are homogenous on gametype.
      SplitTeam($team_name);
      foreach $team_newname (keys %teams_new) {
         # Determine gametypes+enttypes coverage of subteam.
         $coverage_bitfield = CoverageForSubteam($team_newname);
         # Compare that to the labels coverage.
         CheckSubteamLabelCoverage($team_name, $team_newname, $coverage_bitfield);
      }
      # See if there are any labels left over.
      CheckDanglingLabels($team_name);
      %teams_new = ();
      print STDOUT "\n";
   }
}

sub WriteHeader {
   my($map_ref);

   if ($no_write) {
      return;
   }

   $map_ref = substr($map_image, rindex($map_image, "/") + 1);
   $bspname = $map_ref;
   $bspname =~ s/\.[^\.]*$//i;
   $map_ref = "$map_path/".$map_ref;

   if (!open(HTMLFILE, ">$html_dir/$bspname".".html")) {
      die "FATAL: Couldn't write webpage $html_dir/$bspname"."html!";
   }

   print HTMLFILE<<EOB;
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>$bspname : $map_name by $author</title>
<!--
CSSP and form for icon management autogenerated by Entifier version $version, by Joel "Johnny Law" Baxter.
Apologies in advance for stupidities, redundancies, and variations from strict HTML 4.01 required by browser compatibility gymnastics.
-->
<script type="text/javascript" src="iconmanager.js">
</script>
</head>
<body style="background-color: white" onLoad="getVisSyntax();initForm();setGametype(0);show('iconSel');enforceMaxWidth()">
<div id="blackbg" style="position: absolute; left: 0px; top: 0px; z-index: 0; background-color: black; visibility: visible"><img src="page_graphics/blackbg.jpg" width=$map_width height=$map_height alt="" id="blackbgImg"></div>
<div style="position: absolute; left: 0px; top: 0px; z-index: 10000"><img src="page_graphics/frame.gif" width=$map_width height=$map_height alt=""></div>
<div style="position: absolute; left: 672px; top: 472px; z-index: 10001"><img src="$levelshot_path/$bspname.jpg" width=128 height=128 alt=""></div>
<div id="testhidden" style="position: absolute; left: 0px; top: 0px; z-index: 0; visibility: hidden"></div>
<div id="mapview" style="position: absolute; left: 0px; top: 0px; z-index: 1; visibility: visible"><img src="$map_ref" width=$map_width height=$map_height alt="" id="mapviewImg"></div>
EOB
}

sub WriteFooter {
   my($form_pos);
   my($form_width);
   my($ctf_option, $cctf_option);
   my($ta_ctf_option, $ta_1flag_option, $ta_overload_option, $ta_harvester_option);
   my($nav_buttons);

   if ($no_write) {
      return;
   }

   $form_pos = $map_height + 10;
   $form_pos .= "px";
   $form_width = $map_width."px";

   if ($no_ctf) {
      $ctf_option = "<input type=\"hidden\" id=\"setGametypesRadio0\" name=\"setGametypes\">";
   }
   else {
      $ctf_option = "<input type=\"radio\" id=\"setGametypesRadio0\" name=\"setGametypes\" onClick=\"setGametype(0)\">CTF";
   }
   if ($no_cctf) {
      $cctf_option = "<input type=\"hidden\" id=\"setGametypesRadio1\" name=\"setGametypes\">";
   }
   else {
      $cctf_option = "<input type=\"radio\" id=\"setGametypesRadio1\" name=\"setGametypes\" onClick=\"setGametype(1)\">CCTF";
   }
   if ($no_ta) {
      $ta_ctf_option = "<input type=\"hidden\" id=\"setGametypesRadio2\" name=\"setGametypes\">";
      $ta_1flag_option = "<input type=\"hidden\" id=\"setGametypesRadio3\" name=\"setGametypes\">";
      $ta_overload_option = "<input type=\"hidden\" id=\"setGametypesRadio4\" name=\"setGametypes\">";
      $ta_harvester_option = "<input type=\"hidden\" id=\"setGametypesRadio5\" name=\"setGametypes\">";
   }
   else {
      $ta_ctf_option = "<input type=\"radio\" id=\"setGametypesRadio2\" name=\"setGametypes\" onClick=\"setGametype(2)\">TA CTF";
      $ta_1flag_option = "<input type=\"radio\" id=\"setGametypesRadio3\" name=\"setGametypes\" onClick=\"setGametype(3)\">TA Oneflag";
      $ta_overload_option = "<input type=\"radio\" id=\"setGametypesRadio4\" name=\"setGametypes\" onClick=\"setGametype(4)\">TA Overload";
      $ta_harvester_option = "<input type=\"radio\" id=\"setGametypesRadio5\" name=\"setGametypes\" onClick=\"setGametype(5)\">TA Harvester";
   }

   if ($no_nav) {
      $nav_buttons = "";
   }
   else {
      $nav_buttons = "<table border=0 cellpadding=0 cellspacing=0 width=96>\n<colgroup>\n<col width=32>\n<col width=32>\n<col width=32>\n<colgroup>\n<tr><td></td><td><a href=\"$top_page.html\"><img border=0 src=\"page_graphics/up_arrow.gif\" width=32 height=32 alt=\"top\"></a></td><td></td></tr>\n<tr><td><a href=\"$prev_page.html\"><img border=0 src=\"page_graphics/back_arrow.gif\" width=32 height=32 alt=\"prev\"></a></td><td></td><td><a href=\"$next_page.html\"><img border=0 src=\"page_graphics/forward_arrow.gif\" width=32 height=32 alt=\"next\"></a></td></tr>\n</table>";
   }

print HTMLFILE<<EOB;
<div id="iconSel" style="position: absolute; left: 0px; top: $form_pos; width: $form_width; text-align: center; visibility: hidden">
<center>
<form action="">
<input type="checkbox" id="mapCheck" name="map" onClick="toggleMap()">Map
<input type="checkbox" id="armorCheck" name="armor" onClick="toggleType('a')">Armor
<input type="checkbox" id="weaponsCheck" name="weapons" onClick="toggleType('w')">Weapons
<input type="checkbox" id="powerupsCheck" name="powerups" onClick="toggleType('p')">Powerups
<input type="checkbox" id="spawnsCheck" name="spawns" onClick="toggleType('s')">Spawns
<input type="checkbox" id="respawnsCheck" name="respawns" onClick="toggleType('r')">Respawns<BR>
$ctf_option
$cctf_option
$ta_ctf_option
$ta_1flag_option
$ta_overload_option
$ta_harvester_option
</form>
<p>
<b>$bspname : $map_name by $author</b>
</p>
$nav_buttons
</center>
</div>
<noscript>
<div style="position: absolute; left: 0px; top: $form_pos; width: $form_width; text-align: center">
<center>
<p>
You must have JavaScript enabled to control icon display on the map view.
</p>
<p>
<b>$bspname : $map_name by $author</b>
</p>
$nav_buttons
</center>
</noscript>
</body>
</html>
EOB

   close(HTMLFILE);
}

sub CheckOptions {
   if ($ent_file eq "") {
      die "FATAL: Must specify entfile!\n";
   }
   if ($ent_cfg eq "") {
      die "FATAL: Must specify entcfg!\n";
   }
   if (!defined($map_width)) {
      if (!$no_write) {
         die "FATAL: Couldn't read size of image $map_image!\n";
      }
      else {
         $map_width = 800;
         $map_height = 600;
      }
   }
   if (!$no_write) {
      if (($map_width != 800) || ($map_height != 600)) {
         die "FATAL: Map image must be 800x600!\n";
      }
      if ($map_image eq "") {
         die "FATAL: Must specify mapimage!\n";
      }
      if ($author eq "") {
         die "FATAL: Must specify author!\n";
      }
      if (!$no_nav) {
         if ($top_page eq "") {
            die "FATAL: Must specify toppage!\n";
         }
         if ($prev_page eq "") {
            die "FATAL: Must specify prevpage!\n";
         }
         if ($next_page eq "") {
            die "FATAL: Must specify nextpage!\n";
         }
      }
   }
}

sub ShowOptions {
   print "Options:\n";
   print "   mapimage  = $map_image\n";
   print "   author    = $author\n";
   print "   entfile   = $ent_file\n";
   print "   entcfg    = $ent_cfg\n";
   print "   altitude  = $altitude\n";
   print "   yaw       = $yaw\n";
   print "   xshift    = $shift_x\n";
   print "   yshift    = $shift_y\n";
   print "   icondir   = $icon_dir\n";
   print "   htmldir   = $html_dir\n";
   print "   mappath   = $map_path\n";
   print "   thumbpath = $levelshot_path\n";
   print "   iconpath  = $icon_path\n";
   print "   toppage   = $top_page\n";
   print "   prevpage  = $prev_page\n";
   print "   nextpage  = $next_page\n";
   print "\noverhead image width = $map_width, height = $map_height\n";

   if ($no_cctf) {
      print "\n*** hiding CCTF gametype checkbox ***\n";
   }
   if ($no_ta) {
      print "\n*** hiding TA gametype checkbox ***\n";
   }
   if ($no_cctf && $no_ta) {
      $no_ctf = 1;
      print "\n*** hiding CTF gametype checkbox (only gametype checkbox remaining) ***\n";
   }
   if ($no_write) {
      print "\n*** not writing HTML page ***\n";
   }
   else {
      if ($no_nav) {
         print "\n*** writing HTML page without nav buttons ***\n";
      }
   }
}

sub Usage {
   die<<EOB;

Usage: entifier.pl --entfile filepath  [other options]

  --entfile is the only option that absolutely must be specified in all
  circumstances.  "filepath" specifies the path of the entities description
  for the map; this file can either be the .map file itself, or an entities
  file extracted from a .bsp by q3wumpas (or some other entity list of that
  same format).


  --help or --usage will show this message.


  --nonav will omit the navigation buttons at the bottom of the HTML page.


  --nowrite can be specified if you do not want to create an HTML page at
  all; i.e. if you're only interested in seeing the various warning messages.


  *** If --nowrite is NOT specified, and therefore an HTML page is being
  created, then the following options MUST be specified:

  --mapimage filepath, where "filepath" specifies the path of the 800x600
  overhead view graphic.  The root of the graphic's filename must be the same
  as the root of the map's .bsp filename.

  --author name, where "name" is the author of the map.


  *** If --nowrite and --nonav are BOTH not specified, and therefore an HTML
  page is being created with navigation buttons, then the following options
  MUST be specified:

  --prevpage name, where "name" is the name of the page that should be linked
  to the navigation back-arrow.  Not a full path, just the portion of the
  page's filename WITHOUT the .html suffix.

  --nextpage name, as above but for the page to be linked to the navigation
  forward-arrow.


  *** The following are always optional, as they have default values:

  --entcfg filepath, where "filepath" specifies the file that describes the
  properties of entities; default "entity.cfg".

  --altitude units, where "units" is the altitude that was specified when
  creating the overhead view image; default "4000".

  --yaw units, where "units" is the yaw that was specified when creating the
  overhead view image; default "0".

  --xshift units, where "units" is the x shift that was specified when
  creating the overhead view image; default "0".

  --yshift units, where "units" is the y shift that was specified when
  creating the overhead view image; default "0".

  --icondir dir, where "dir" is the directory containing the icon graphics;
  default "entity_icons".

  --htmldir dir, where "dir" is the directory where HTML page output should
  be writting, default ".".

  --mappath path, where "path" is the relative URL (relative the location of
  the HTML page) where the overhead view graphics will be stored when viewing
  the page; default "overheads".  The graphic must have the same filename as
  the overhead view graphic used when generating the HTML page.

  --thumbpath path, where "path" is the relative URL where the thumbnail
  images (levelshots) of maps will be stored when viewing the page; default
  "levelshots".  The thumbnail/levelshot for a map must be a .jpg image,
  with the root of the filename the same as the root of the overhead graphic's
  filename.

  --iconpath path, where "path" is the relative URL where the icon graphics
  will be stored when viewing the page; default "entity_icons".

  --toppage name, like prevpage or nextpage but for the page to be linked
  to the navigation up-arrow; default "index".


  *** And there are two final optional switches for maps that don't support
  all the gametypes:

  --nocctf, specified if the map has no entity placement specific to the
  CCTF gametype.

  --nota, specified if the map has no entity placement specific to the
  Team Arena gametypes.
EOB
}


if (!GetOptions('mapimage=s' => \$map_image,
                'author=s'   => \$author,
                'entfile=s'  => \$ent_file,
                'entcfg=s'   => \$ent_cfg,
                'altitude=f' => \$altitude,
                'yaw=f'      => \$yaw,
                'xshift=f'   => \$shift_x,
                'yshift=f'   => \$shift_y,
                'icondir=s'  => \$icon_dir,
                'htmldir=s'  => \$html_dir,
                'mappath=s'  => \$map_path,
                'thumbpath=s'=> \$levelshot_path,
                'iconpath=s' => \$icon_path,
                'toppage=s'  => \$top_page,
                'prevpage=s' => \$prev_page,
                'nextpage=s' => \$next_page,
                'nocctf'     => \$no_cctf,
                'nota'       => \$no_ta,
                'nonav'      => \$no_nav,
                'nowrite'    => \$no_write,
                'help'       => \$help,
                'usage'      => \$help)) { Usage(); }

if ($help) {
   Usage();
}

($map_width, $map_height) = imgsize($map_image);

CheckOptions();

ShowOptions();

InitEntities();

InitIcons();

WriteHeader();

PlaceEntities();

WriteFooter();

AnalyzeTeams();
